#!/usr/bin/python3

"""
Checks perf test results for regressions.
Usage: ./script_name.py <pattern> <threshold>
"""

import glob
import os
import sys


def print_usage():
    print(f"Usage: {sys.argv[0]} <pattern> <threshold>")
    print("\t<pattern>: Glob pattern to match performance test files.")
    print("\t<threshold>: Threshold to detect significant regression.")
    exit(1)


def check_perf_results(pattern, threshold):
    max_regression = 0
    test_name_max = file_name_max = None

    for filename in glob.glob(pattern):
        with open(filename, "r") as f:
            test_name = None
            for line in f:
                if "+osu" in line:
                    test_name = line.split()[0].strip("+")
                elif "% worse" in line:
                    worse_val = float(line.split("%")[0].split()[-1])
                    if worse_val > max_regression:
                        max_regression = worse_val
                        test_name_max = test_name
                        file_name_max = os.path.basename(filename)

    if max_regression > threshold:
        msg = f"Max regression in {test_name_max} from {file_name_max}: {max_regression}%"
        print(msg)
        print(f"##vso[task.complete result=SucceededWithIssues;]DONE{msg}")
    else:
        print("No significant performance regression detected.")


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print_usage()
    check_perf_results(sys.argv[1], float(sys.argv[2]))



#!/usr/bin/env python3
#
# Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2017. ALL RIGHTS RESERVED.
#
# See file LICENSE for terms.
#

import sys
import subprocess
import os
import re
import itertools
import contextlib
from optparse import OptionParser
from pkg_resources import parse_version


#expected AM transport selections per given number of eps
mlx4_am = {
       2 :      "rc_verbs",
      16 :      "rc_verbs",
      32 :      "rc_verbs",
      64 :      "rc_verbs",
     256 :      "ud_verbs",
     512 :      "ud_verbs",
    1024 :      "ud_verbs",
 1000000 :      "ud_verbs",
}

mlx5_am = {
       2 :      "rc_mlx5",
      16 :      "rc_mlx5",
      32 :      "rc_mlx5",
      64 :      "dc_mlx5",
     256 :      "dc_mlx5",
     512 :      "dc_mlx5",
    1024 :      "dc_mlx5",
 1000000 :      "dc_mlx5",
}

mlx5_am_no_dc = {
       2 :      "rc_mlx5",
      16 :      "rc_mlx5",
      32 :      "rc_mlx5",
      64 :      "rc_mlx5",
     256 :      "ud_mlx5",
     512 :      "ud_mlx5",
    1024 :      "ud_mlx5",
 1000000 :      "ud_mlx5",
}

# check that UCX_NUM_EPS work
mlx5_am_override = {
       2 :      "rc_mlx5",
      16 :      "rc_mlx5",
      32 :      "rc_mlx5",
      64 :      "rc_mlx5",
     256 :      "rc_mlx5",
     512 :      "rc_mlx5",
    1024 :      "rc_mlx5",
 1000000 :      "rc_mlx5",
}

mlx4_am_override = {
       2 :      "rc_verbs",
      16 :      "rc_verbs",
      32 :      "rc_verbs",
      64 :      "rc_verbs",
     256 :      "rc_verbs",
     512 :      "rc_verbs",
    1024 :      "rc_verbs",
 1000000 :      "rc_verbs",
}

am_tls =  {
    "mlx4"            : mlx4_am,
    "mlx5"            : mlx5_am,
    "mlx5_roce_dc"    : mlx5_am,       # mlx5 RoCE port which supports DC
    "mlx5_roce_no_dc" : mlx5_am_no_dc, # mlx5 RoCE port which doesn't support DC
    "mlx4_override"   : mlx4_am_override,
    "mlx5_override"   : mlx5_am_override
}

tl_aliases = {
    "mm":   ["posix", "sysv", "xpmem", ],
    "sm":   ["posix", "sysv", "xpmem", "knem", "cma", "rdmacm", "sockcm", ],
    "shm":  ["posix", "sysv", "xpmem", "knem", "cma", "rdmacm", "sockcm", ],
    "ib":   ["rc_verbs", "ud_verbs", "rc_mlx5", "ud_mlx5", "dc_mlx5", "rdmacm",
             "ud_mlx5:aux", "ud_verbs:aux", ],
    "ud_v": ["ud_verbs", "rdmacm", ],
    "ud_x": ["ud_mlx5", "rdmacm", ],
    "ud":   ["ud_mlx5", "ud_verbs", "rdmacm", ],
    "rc_v": ["rc_verbs", "ud_verbs:aux", "rdmacm", ],
    "rc_x": ["rc_mlx5", "ud_mlx5:aux", "rdmacm", ],
    "rc":   ["rc_mlx5", "ud_mlx5:aux", "rc_verbs", "ud_verbs:aux", "rdmacm", ],
    "dc":   ["dc_mlx5", "rdmacm", ],
    "dc_x": ["dc_mlx5", "rdmacm", ],
    "ugni": ["ugni_smsg", "ugni_udt:aux", "ugni_rdma", ],
    "cuda": ["cuda_copy", "cuda_ipc", "gdr_copy", ],
    "rocm": ["rocm_copy", "rocm_ipc", "rocm_gdr", ],
}

@contextlib.contextmanager
def _override_env(env_vars):
    prev_values = []
    for var_name, value in env_vars:
        prev_values.append((var_name, os.getenv(var_name)))
        if value is not None:
            os.putenv(var_name, value)
        else:
            os.unsetenv(var_name)

    try:
        yield
    finally:
        for var_name, prev_value in prev_values:
            os.putenv(var_name, prev_value) if prev_value else os.unsetenv(var_name)

def exec_cmd(cmd):
    if options.verbose:
        print(cmd)

    status, output = subprocess.getstatusoutput(cmd)
    if options.verbose:
        print(f"return code {status}")
        print(output)

    return status, output

def find_transport(dev=None, neps=1, override=0, tls="ib", protocol="am"):
    if (override):
        os.putenv("UCX_NUM_EPS", "2")

    env_vars = [("UCX_TLS", tls)]

    # Set up environment variables based on protocol type
    if protocol == "am" and dev:
        env_vars.append(("UCX_NET_DEVICES", dev))

    # Use context manager for all environment variables
    with _override_env(env_vars):
        # Choose the appropriate arguments and grep pattern based on protocol type
        if protocol == "keepalive":
            args = ucx_info_eh_args
        elif protocol == "am":  # am transport
            args = ucx_info_args

        status, output = exec_cmd(f"{ucx_info}{args}{neps} | grep {protocol}")

    match = re.search(r'\d+:(\S+)/\S+', output)
    if match:
        proto_tls = match.group(1)
        if override:
            os.unsetenv("UCX_NUM_EPS")

        return proto_tls
    else:
        return None

def find_am_transport(dev, neps=1, override=0, tls="ib"):
    return find_transport(dev=dev, neps=neps, override=override,
                          tls=tls, protocol="am")

def test_fallback_from_rc(dev, neps) :

    os.putenv("UCX_TLS", "ib")
    os.putenv("UCX_NET_DEVICES", dev)

    status, output = exec_cmd(f"{ucx_info}{ucx_info_args}{neps} | grep rc")

    os.unsetenv("UCX_TLS")
    os.unsetenv("UCX_NET_DEVICES")

    if output != "":
        print(f"RC transport must not be used when estimated number of EPs = {neps}")
        sys.exit(1)

    os.putenv("UCX_TLS", "rc,ud,tcp")

    status, output_rc = exec_cmd(f"{ucx_info}{ucx_info_args}{neps} | grep rc")

    status, output_tcp = exec_cmd(f"{ucx_info}{ucx_info_args}{neps} | grep tcp")

    if output_rc != "" or output_tcp != "":
        print(f"RC/TCP transports must not be used when estimated number of EPs = {neps}")
        sys.exit(1)

    os.unsetenv("UCX_TLS")

def test_ucx_tls_positive(tls):
    # Use TLS list in "allow" mode and verify that the found tl is in the list
    found_tl = find_am_transport(None, tls=tls)
    print(f"Using UCX_TLS={tls}, found TL: {found_tl}")
    if tls == 'all':
        return
    if not found_tl:
        sys.exit(1)
    tls = tls.split(',')
    if found_tl in tls or f"\\{found_tl}" in tls:
        return
    for tl in tls:
        if tl in tl_aliases and found_tl in tl_aliases[tl]:
            return
    print("Found TL doesn't belong to the allowed UCX_TLS")
    sys.exit(1)

def test_ucx_tls_negative(tls, protocol="am", forbidden_tls=None):
    # Use TLS list in "negate" mode
    found_tl = find_transport(tls="^"+tls, protocol=protocol)
    print(f"Using UCX_TLS=^{tls}, found {protocol} TL: {found_tl}")
    if not found_tl:
        print("No available TL found")
        sys.exit(1)

    # If forbidden_tls is provided, verify that the found tl is not in that list
    if forbidden_tls is not None:
        if found_tl in forbidden_tls:
            print(f"Found forbidden TL: {found_tl}")
            sys.exit(1)
        return

    # Otherwise, check against the tls list
    tls = tls.split(',')
    if found_tl in tls:
        print(f"Found forbidden TL: {found_tl}")
        sys.exit(1)
    for tl in tls:
        if tl in tl_aliases and found_tl in tl_aliases[tl]:
            print(f"Found forbidden TL: {found_tl}")
            sys.exit(1)

def _powerset(iterable, with_empty_set=True):
    iterable_list = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(iterable_list, r) for r in \
            range(0 if with_empty_set else 1, len(iterable_list) + 1))

def test_tls_allow_list(ucx_info):
    status, output = exec_cmd(ucx_info + " -d | grep Transport | awk '{print $3}'")
    available_tls = set(output.splitlines())

    # Add some basic variants (those that are available on this platform)
    tls_variants = [tls_variant for tls_variant in ["tcp", "posix", "xpmem"] if \
                    tls_variant in available_tls]

    # Add some IB variant (both strict and alias), if available
    for tls_variant in available_tls:
        if tls_variant.startswith("dc_") or tls_variant.startswith("ud_"):
            tls_variants += ["ib", "\\" + tls_variant]
            break

    tls_variants = _powerset(tls_variants, with_empty_set=False)
    test_funcs = [test_ucx_tls_positive, test_ucx_tls_negative]
    for (tls_variant, test_func) in \
        itertools.product(tls_variants, test_funcs):
        test_func(",".join(tls_variant))

    # Test auxiliary transport negation
    test_cases_negative = [
        ("ib", {"ud_mlx5", "ud_verbs"}),
        ("ud,ud:aux", {"ud_mlx5", "ud_verbs"}),
        ("ud_v,ud_v:aux", {"ud_verbs"}),
        ("ud_x,ud_x:aux", {"ud_mlx5"}),
        ("ud_verbs,ud_verbs:aux", {"ud_verbs"}),
        ("ud_mlx5,ud_mlx5:aux", {"ud_mlx5"})
    ]

    for tls, forbidden_tls in test_cases_negative:
        test_ucx_tls_negative(tls, protocol="keepalive", forbidden_tls=forbidden_tls)

parser = OptionParser()
parser.add_option("-p", "--prefix", metavar="PATH", help = "root UCX directory")
parser.add_option("-v", "--verbose", action="store_true", \
                  help = "verbose output", default=False)
(options, args) = parser.parse_args()

if options.prefix == None:
    bin_prefix = "./src/tools/info"
else:
    bin_prefix = options.prefix + "/bin"

if not (os.path.isdir(bin_prefix)):
    print(f"directory \"{bin_prefix}\" does not exist")
    parser.print_help()
    exit(1)

ucx_info = bin_prefix + "/ucx_info"
ucx_info_args = " -e -u t -n "
ucx_info_eh_args = " -e -u et -n "

status, output = exec_cmd(ucx_info + " -c | grep -e \"UCX_RC_.*_MAX_NUM_EPS\"")
match = re.findall(r'\S+=(\d+)', output)
if match:
    rc_max_num_eps = int(max(match))
else:
    rc_max_num_eps = 0

status, output = exec_cmd("ibv_devinfo -l | tail -n+2 | head -n-1 | sed -e 's/^[ \t]*//' | grep -v '^smi[0-9]*$'")
dev_list = output.splitlines()
port = "1"

for dev in sorted(dev_list):
    status, dev_attrs = exec_cmd("ibv_devinfo -d " + dev + " -i " + port)
    if dev_attrs.find("PORT_ACTIVE") == -1:
        continue

    if not os.path.exists(f"/sys/class/infiniband/{dev}/ports/{port}/gids/0"):
        print("Skipping dummy device: ", dev)
        continue

    driver_name = os.path.basename(os.readlink(f"/sys/class/infiniband/{dev}/device/driver"))
    dev_name    = driver_name.split("_")[0] # should be mlx4 or mlx5
    if not dev_name in ['mlx4', 'mlx5']:
        print("Skipping unknown device: ", dev_name)
        continue

    if dev_attrs.find("Ethernet") == -1:
        dev_tl_map = am_tls[dev_name]
        dev_tl_override_map = am_tls[dev_name + "_override"]
        override = 1
    else:
        fw_ver = open(f"/sys/class/infiniband/{dev}/fw_ver").read()
        if parse_version(fw_ver) >= parse_version("16.23.0"):
            dev_tl_map = am_tls[dev_name+"_roce_dc"]
        else:
            dev_tl_map = am_tls[dev_name+"_roce_no_dc"]
        override = 0

    for n_eps in sorted(dev_tl_map):
        tl = find_am_transport(dev + ':' + port, n_eps)
        print(f"{dev}:{port} eps: {n_eps} expected am tl: {dev_tl_map[n_eps]} selected: {tl}")

        if dev_tl_map[n_eps] != tl:
            sys.exit(1)

        if override:
            tl = find_am_transport(f"{dev}:{port}", n_eps, 1)
            print(f"{dev}:{port} UCX_NUM_EPS=2 eps: {n_eps} expected am tl: \
                  {dev_tl_override_map[n_eps]} selected: {tl}")

            if dev_tl_override_map[n_eps] != tl:
                sys.exit(1)

        if n_eps >= (rc_max_num_eps * 2):
            test_fallback_from_rc(f"{dev}:{port}", n_eps)

# Test UCX_TLS configuration (TL choice according to "allow" and "negate" lists)
test_tls_allow_list(ucx_info)

sys.exit(0)




# -*- coding: utf-8 -*-
#
# Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2019. ALL RIGHTS RESERVED.
# Copyright (C) Arm Ltd. 2021.  ALL RIGHTS RESERVED.
#
# See file LICENSE for terms.
#

#
# Project information
#

project   = u'OpenUCX'
copyright = u'2019, UCF'
author    = u'UCF'

#
# General options
#

extensions         = ['recommonmark', 'breathe'] # For processing Markdown pages
templates_path     = ['_templates']
source_suffix      = ['.rst', '.md']
master_doc         = 'index'
language           = None
exclude_patterns   = [u'_build']
pygments_style     = None

#
# Breathe configuration
#

# breathe_separate_member_pages      = False
# breathe_show_enumvalue_initializer = False
# breathe_show_define_initializer    = False
breathe_order_parameters_first     = True
breathe_projects = {"openucx":"build/xml/",}

#                  
# HTML options     
#                  

html_theme         = 'sphinx_rtd_theme'
html_logo          = '_static/ucxlogo.png'
html_theme_options = {
    'style_external_links': True
}
html_static_path   = ['_static']
htmlhelp_basename  = 'OpenUCXdoc'


#
# UCX custom configuration
#

# For the local testing set cached_version to the latest released version (e.g 'v1.10.1').
# Otherwise, GitHub may block you for a frequent access to GitHub API (DDOS protection).
cached_version = ''

def getLatestVersion():
    import re
    import os

    global cached_version
    if cached_version == '':
        try:
            with open(os.path.join(os.path.dirname(__file__), '../../configure.ac'), "r") as f:
                content = f.read()
                major = re.search(r'define\(\[ucx_ver_major\],\s*(\d+)\)', content).group(1)
                minor = re.search(r'define\(\[ucx_ver_minor\],\s*(\d+)\)', content).group(1)
                patch = re.search(r'define\(\[ucx_ver_patch\],\s*(\d+)\)', content).group(1)
                cached_version = "v{}.{}.{}".format(major, minor, patch)
        except Exception as e:
            raise RuntimeError("Failed to retrieve the version from configure.ac: " + str(e))

    return cached_version

def substituteVersion(app, docname, source):
    #
    # Updating these 2 variables will automatically update all download and API
    # documentation links.
    # We don't use the normal RST substitution because it cannot substitute text
    # inside code blocks and URL links.
    #
    version_name  = getLatestVersion()
    clean_version = version_name.lstrip('v')        # remove leading 'v' for tag name
    api_version   = clean_version.rsplit('.', 1)[0] # take only MAJOR.MINOR
    result        = source[0].replace("{VERSION}", api_version) \
                             .replace("{RELEASE}", clean_version)
    source[0] = result

def run_doxygen(doxygen_home):
    import subprocess, sys

    try:
        ret = subprocess.call('cd %s && doxygen && cd -' % doxygen_home, shell=True)
        if ret < 0:
            sys.stderr.write('doxygen error code %s' % (-ret))
    except OSError as e:
        sys.stderr.write('doxygen execution failed: %s' % e)

def run_apidoc(api_dir, xml_dir):
    import subprocess, sys

    try:
        # We don't generate groups since we create those manually
        ret = subprocess.call('breathe-apidoc -m -o %s -p openucx %s -g struct,file' % (api_dir, xml_dir), shell=True)
        if ret < 0:
            sys.stderr.write('breathe-apidoc error code %s' % (-ret))
    except OSError as e:
        sys.stderr.write('breathe-apidoc execution failed: %s' % e)

def config_doxygen_file(dox_dir, src_dir, output_dir):
    with open(dox_dir +'/ucxdox', 'r') as file :
        doxygenfile = file.read()
    # Using upstream doxygen configuration to generate "breathe" friendly
    # configuration. Since autotools are not available at RTD we are doing this
    # python way
    doxygenfile = doxygenfile.replace('$(PROJECT)', 'UCX')
    doxygenfile = doxygenfile.replace('$(VERSION)', getLatestVersion())
    doxygenfile = doxygenfile.replace('$(SRCDIR)', src_dir)
    doxygenfile = doxygenfile.replace('$(DOCDIR)', output_dir)
    doxygenfile = doxygenfile.replace('$(GENERATE_HTML)', 'NO')
    doxygenfile = doxygenfile.replace('$(GENERATE_HTMLHELP)', 'NO')
    doxygenfile = doxygenfile.replace('$(GENERATE_CHI)', 'NO')
    doxygenfile = doxygenfile.replace('$(GENERATE_LATEX)', 'NO')
    doxygenfile = doxygenfile.replace('$(GENERATE_RTF)', 'NO')
    doxygenfile = doxygenfile.replace('$(GENERATE_MAN)', 'NO')
    doxygenfile = doxygenfile.replace('$(GENERATE_XML)', 'YES')
    doxygenfile = doxygenfile.replace('$(PERL_PATH)', '/usr/bin/perl')
    doxygenfile = doxygenfile.replace('$(HAVE_DOT)', 'NO')
    doxygenfile = doxygenfile.replace('docs/doxygen/header.tex', '')
    # Exclude UCT documentation
    doxygenfile = doxygenfile.replace('../..//src/uct/api/', '')

    with open(dox_dir + '/Doxyfile', 'w') as file:
        file.write(doxygenfile)

def generate_api(app):
    import os

    read_the_docs_build = os.environ.get('READTHEDOCS', None) == 'True'
    if read_the_docs_build:
        dox_dir = '../doxygen'
        api_dir = 'api'
        xml_dir = 'build/xml/'
    else:
        dox_dir = 'doxygen'
        api_dir = 'source/api'
        xml_dir = 'source/build/xml/'

    output_dir = '../source/build'
    src_dir = '../../'

    config_doxygen_file(dox_dir, src_dir, output_dir)
    run_doxygen(dox_dir)
    run_apidoc(api_dir, xml_dir)

def setup(app):
    app.connect('builder-inited', generate_api)
    app.connect('source-read', substituteVersion)



#!/usr/bin/env python3

#
# Copyright (C) 2022 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# See file LICENSE for terms.

import os
import re
import argparse
import subprocess
import sys
import logging


class Environment(object):
    '''Handles environment variables setup and cleanup'''
    def __init__(self, env_vars):
        logging.info(f'Using env vars: {env_vars}')
        self.env_vars = env_vars;

    def __enter__(self):
        self.cleanup()
        for var_name in self.env_vars:
            os.environ[var_name] = 'value'
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def cleanup(self):
        ucx_vars = [var for var in os.environ.keys() if var.startswith('UCX_')]
        for var in ucx_vars:
            del os.environ[var]


class TestRunner:
    '''Main test runner'''
    def __init__(self, ucx_info, verbose):
        self.ucx_info = ucx_info
        if verbose:
            logging.basicConfig(level=logging.DEBUG)

    def run(self, test_case):
        with Environment(test_case.keys()):
            matches = self.get_fuzzy_matches()

            if matches != test_case:
                raise Exception(f'Wrong fuzzy list: got: {matches}, expected: {test_case}')

            logging.info(f'found all expected matches: {test_case}')

    def exec_ucx_info(self):
        cmd = f'{self.ucx_info} -u m -w'
        logging.info(f'running cmd: {cmd}')

        status, output = subprocess.getstatusoutput(cmd)
        if status != 0:
            raise Exception(f'Received unexpected exit code from ucx_info: {status}')

        logging.info(output)
        return output

    def get_fuzzy_matches(self):
        output = self.exec_ucx_info()
        output_lines = output.splitlines()
        warn_msg = output_lines[0]

        # This text is printed from 'parser.c' file (updates should be synced properly).
        warn_match = re.match('.*unused environment variables?: (.*)', warn_msg)
        if not warn_match:
            print("First 5 lines of output:")
            print("\n".join(output_lines[:5]))
            raise Exception('"unused vars" message was not found')

        output_vars = warn_match.group(1).split(';')
        matches = [re.match(r'(\w+)(?: \(maybe: (.*)\?\))?', var.strip()) for var in output_vars]
        if None in matches:
            raise Exception(f'Unexpected warning message format: {warn_msg}')

        return {m.group(1) : [x.strip() for x in m.group(2).split(',')] if m.group(2) else [] for m in matches}

def has_ib():
    status, output = subprocess.getstatusoutput('ibv_devinfo')
    if status != 0:
        return False

    return 'No IB devices found' not in output

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Tester for config vars fuzzy matching')
    parser.add_argument('--ucx_info', help="Path to ucx_info binary", required=True)
    parser.add_argument('-v', '--verbose', help="Add some debug prints", action='store_true', default=False)
    args = parser.parse_args()

    try:
        runner = TestRunner(args.ucx_info, args.verbose)
        test_cases =  [{'UCX_LOF_LEVEL' : ['UCX_LOG_LEVEL']},
                       {'UCX_LOF_LEVEL' : ['UCX_LOG_LEVEL'], 'UCX_MOFULE_D' : ['UCX_MODULE_DIR', 'UCX_MODULES']},
                       {'UCX_SOME_VAR' : [], 'UCX_SOME_VAR2' : [],  'UCX_SOME_VAR3' : [],  'UCX_SOME_VAR4' : []},
                       {'UCX_SOME_VAR' : [], 'UCX_MOFULE_D' : ['UCX_MODULE_DIR', 'UCX_MODULES'], 'UCX_SOME_VAR2' : [], 'UCX_LOF_LEVEL' : ['UCX_LOG_LEVEL']},
                       {'UCX_RLS' : ['UCX_TLS']}]

        if has_ib():
            test_cases += [{'UCX_RC_VERBS_RX_MAX_BUF' : ['UCX_RC_VERBS_TX_MAX_BUFS', 'UCX_RC_VERBS_RX_MAX_BUFS', 'UCX_UD_VERBS_RX_MAX_BUFS']}]

        for test_case in test_cases:
            runner.run(test_case)

    except Exception as e:
        logging.error(str(e))
        sys.exit(1)
